/* -*- mode: c; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4; c-file-style: "stroustrup"; -*-
 *
 *
 *                This source code is part of
 *
 *                 G   R   O   M   A   C   S
 *
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2009, The GROMACS Development Team
 *
 * Gromacs is a library for molecular simulation and trajectory analysis,
 * written by Erik Lindahl, David van der Spoel, Berk Hess, and others - for
 * a full list of developers and information, check out http://www.gromacs.org
 *
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU Lesser General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) any 
 * later version.
 * As a special exception, you may use this file as part of a free software
 * library without restriction.  Specifically, if other files instantiate
 * templates or use macros or inline functions from this file, or you compile
 * this file and link it with other files to produce an executable, this
 * file does not by itself cause the resulting executable to be covered by
 * the GNU Lesser General Public License.  
 *
 * In plain-speak: do not worry about classes/macros/templates either - only
 * changes to the library have to be LGPL, not an application linking with it.
 *
 * To help fund GROMACS development, we humbly ask that you cite
 * the papers people have written on it - you can find them on the website!
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>

#include "typedefs.h"
#include "vec.h"
#include "smalloc.h"
#include "force.h"
#include "gmx_omp_nthreads.h"
#include "nbnxn_kernel_sse.h"

/* Include all flavors of the SSE kernel loops */

/* Analytical reaction-field kernels */
#define CALC_COUL_RF

/* Include the force+energy kernels */
#define CALC_ENERGIES
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef CALC_ENERGIES

/* Include the force+energygroups kernels */
#define CALC_ENERGIES
#define ENERGY_GROUPS
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef ENERGY_GROUPS
#undef CALC_ENERGIES

/* Include the force only kernels */
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"

#undef CALC_COUL_RF


/* Tabulated Coulomb kernels */

/* Include the force+energy kernels */
#define CALC_ENERGIES
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef CALC_ENERGIES

/* Include the force+energygroups kernels */
#define CALC_ENERGIES
#define ENERGY_GROUPS
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef ENERGY_GROUPS
#undef CALC_ENERGIES

/* Include the force only kernels */
#define LJ_COMB_GEOM
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_GEOM
#define LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"
#undef LJ_COMB_LB
#include "nbnxn_kernel_sse_outer.h"


static void
nbnxn_kernel_sse_single_tab_comb_geom_noener_noLJ_stripped
# 133 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
                            (const nbnxn_pairlist_t *nbl,
                             const nbnxn_atomdata_t *nbat,
                             const interaction_const_t *ic,
                             rvec *shift_vec,
                             real *f,
                             real *fshift)
{
    const nbnxn_ci_t *nbln;
    const nbnxn_cj_t *l_cj;
    const int *type;
    const real *q;
    const real *shiftvec;
    const real *x;
    const real *nbfp0,*nbfp1,*nbfp2=((void *)0),*nbfp3=((void *)0);
    real facel;
    real *nbfp_ptr;
    int nbfp_stride;
    int n,ci,ci_sh;
    int ish,ish3;
    gmx_bool half_LJ,do_coul;
    int sci,scix,sciy,sciz,sci2;
    int cjind0,cjind1,cjind;
    int ip,jp;
# 175 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    __m128 shX_SSE;
    __m128 shY_SSE;
    __m128 shZ_SSE;
    __m128 ix_SSE0,iy_SSE0,iz_SSE0;
    __m128 ix_SSE1,iy_SSE1,iz_SSE1;
    __m128 ix_SSE2,iy_SSE2,iz_SSE2;
    __m128 ix_SSE3,iy_SSE3,iz_SSE3;
    __m128 fix_SSE0,fiy_SSE0,fiz_SSE0;
    __m128 fix_SSE1,fiy_SSE1,fiz_SSE1;
    __m128 fix_SSE2,fiy_SSE2,fiz_SSE2;
    __m128 fix_SSE3,fiy_SSE3,fiz_SSE3;

    __m128 fix_SSE,fiy_SSE,fiz_SSE;
# 199 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    __m128i mask0 = _mm_set_epi32( 0x0008, 0x0004, 0x0002, 0x0001 );
    __m128i mask1 = _mm_set_epi32( 0x0080, 0x0040, 0x0020, 0x0010 );
    __m128i mask2 = _mm_set_epi32( 0x0800, 0x0400, 0x0200, 0x0100 );
    __m128i mask3 = _mm_set_epi32( 0x8000, 0x4000, 0x2000, 0x1000 );
# 225 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    __m128 diag_SSE0 = gmx_mm_castsi128_ps( _mm_set_epi32( 0xffffffff, 0xffffffff, 0xffffffff, 0x00000000 ));
    __m128 diag_SSE1 = gmx_mm_castsi128_ps( _mm_set_epi32( 0xffffffff, 0xffffffff, 0x00000000, 0x00000000 ));
    __m128 diag_SSE2 = gmx_mm_castsi128_ps( _mm_set_epi32( 0xffffffff, 0x00000000, 0x00000000, 0x00000000 ));
    __m128 diag_SSE3 = gmx_mm_castsi128_ps( _mm_set_epi32( 0x00000000, 0x00000000, 0x00000000, 0x00000000 ));
# 258 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    __m128i zero_SSE = _mm_setzero_si128();

    __m128 one_SSE=_mm_set1_ps(1.0);
    __m128 iq_SSE0=_mm_setzero_ps();
    __m128 iq_SSE1=_mm_setzero_ps();
    __m128 iq_SSE2=_mm_setzero_ps();
    __m128 iq_SSE3=_mm_setzero_ps();
    __m128 mrc_3_SSE;


    __m128 invtsp_SSE;
    const real *tab_coul_F;

    int ti0_array[2*4 -1],*ti0;
    int ti1_array[2*4 -1],*ti1;
    int ti2_array[2*4 -1],*ti2;
    int ti3_array[2*4 -1],*ti3;
# 307 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    const real *ljc;

    __m128 c6s_SSE0,c12s_SSE0;
    __m128 c6s_SSE1,c12s_SSE1;
    __m128 c6s_SSE2=_mm_setzero_ps(),c12s_SSE2=_mm_setzero_ps();
    __m128 c6s_SSE3=_mm_setzero_ps(),c12s_SSE3=_mm_setzero_ps();


    __m128 vctotSSE,VvdwtotSSE;
    __m128 sixthSSE,twelvethSSE;

    __m128 rc2_SSE;

    real shf_array[15],*shf;

    int ninner;

    ljc = nbat->lj_comb;
# 350 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    ti0 = (int *)(((size_t)(ti0_array+4 -1)) & (~((size_t)(4*sizeof(real)-1))));
    ti1 = (int *)(((size_t)(ti1_array+4 -1)) & (~((size_t)(4*sizeof(real)-1))));
    ti2 = (int *)(((size_t)(ti2_array+4 -1)) & (~((size_t)(4*sizeof(real)-1))));
    ti3 = (int *)(((size_t)(ti3_array+4 -1)) & (~((size_t)(4*sizeof(real)-1))));

    invtsp_SSE = _mm_set1_ps(ic->tabq_scale);


    tab_coul_F = ic->tabq_coul_FDV0;


    q = nbat->q;
    type = nbat->type;
    facel = ic->epsfac;
    shiftvec = shift_vec[0];
    x = nbat->x;

    rc2_SSE = _mm_set1_ps(ic->rvdw*ic->rvdw);
# 386 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    mrc_3_SSE = _mm_set1_ps(-2*ic->k_rf);
# 398 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    shf = (real *)(((size_t)(shf_array+7)) & (~((size_t)31)));
# 438 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    l_cj = nbl->cj;

    ninner = 0;
    for(n=0; n<nbl->nci; n++)
    {
        nbln = &nbl->ci[n];

        ish = (nbln->shift & 127);
        ish3 = ish*3;
        cjind0 = nbln->cj_ind_start;
        cjind1 = nbln->cj_ind_end;

        ci = nbln->ci;
        ci_sh = (ish == (((2*1 +1)*(2*1 +1)*(2*2 +1))/2) ? ci : -1);

        shX_SSE = _mm_load1_ps(shiftvec+ish3);
        shY_SSE = _mm_load1_ps(shiftvec+ish3+1);
        shZ_SSE = _mm_load1_ps(shiftvec+ish3+2);


        sci = ci*4;
        scix = sci*3;
        sci2 = sci*2;


        half_LJ = (nbln->shift & (1<<(8+3*(0))));
        do_coul = (nbln->shift & (1<<(9+3*(0))));
# 517 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
        sciy = scix + 4;
        sciz = sciy + 4;
        ix_SSE0 = _mm_add_ps(_mm_load1_ps(x+scix) ,shX_SSE);
        ix_SSE1 = _mm_add_ps(_mm_load1_ps(x+scix+1),shX_SSE);
        ix_SSE2 = _mm_add_ps(_mm_load1_ps(x+scix+2),shX_SSE);
        ix_SSE3 = _mm_add_ps(_mm_load1_ps(x+scix+3),shX_SSE);
        iy_SSE0 = _mm_add_ps(_mm_load1_ps(x+sciy) ,shY_SSE);
        iy_SSE1 = _mm_add_ps(_mm_load1_ps(x+sciy+1),shY_SSE);
        iy_SSE2 = _mm_add_ps(_mm_load1_ps(x+sciy+2),shY_SSE);
        iy_SSE3 = _mm_add_ps(_mm_load1_ps(x+sciy+3),shY_SSE);
        iz_SSE0 = _mm_add_ps(_mm_load1_ps(x+sciz) ,shZ_SSE);
        iz_SSE1 = _mm_add_ps(_mm_load1_ps(x+sciz+1),shZ_SSE);
        iz_SSE2 = _mm_add_ps(_mm_load1_ps(x+sciz+2),shZ_SSE);
        iz_SSE3 = _mm_add_ps(_mm_load1_ps(x+sciz+3),shZ_SSE);


        if (do_coul || half_LJ)
        {
            iq_SSE0 = _mm_set1_ps(facel*q[sci]);
            iq_SSE1 = _mm_set1_ps(facel*q[sci+1]);
            iq_SSE2 = _mm_set1_ps(facel*q[sci+2]);
            iq_SSE3 = _mm_set1_ps(facel*q[sci+3]);
        }
# 552 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
        c6s_SSE0 = _mm_load1_ps(ljc+sci2+0);
        c6s_SSE1 = _mm_load1_ps(ljc+sci2+1);
        if (!half_LJ)
        {
            c6s_SSE2 = _mm_load1_ps(ljc+sci2+2);
            c6s_SSE3 = _mm_load1_ps(ljc+sci2+3);
        }
        c12s_SSE0 = _mm_load1_ps(ljc+sci2+4 +0);
        c12s_SSE1 = _mm_load1_ps(ljc+sci2+4 +1);
        if (!half_LJ)
        {
            c12s_SSE2 = _mm_load1_ps(ljc+sci2+4 +2);
            c12s_SSE3 = _mm_load1_ps(ljc+sci2+4 +3);
        }
# 578 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
        VvdwtotSSE = _mm_setzero_ps();
        vctotSSE = _mm_setzero_ps();


        fix_SSE0 = _mm_setzero_ps();
        fix_SSE1 = _mm_setzero_ps();
        fix_SSE2 = _mm_setzero_ps();
        fix_SSE3 = _mm_setzero_ps();
        fiy_SSE0 = _mm_setzero_ps();
        fiy_SSE1 = _mm_setzero_ps();
        fiy_SSE2 = _mm_setzero_ps();
        fiy_SSE3 = _mm_setzero_ps();
        fiz_SSE0 = _mm_setzero_ps();
        fiz_SSE1 = _mm_setzero_ps();
        fiz_SSE2 = _mm_setzero_ps();
        fiz_SSE3 = _mm_setzero_ps();

        cjind = cjind0;


        if (half_LJ)
        {
# 613 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
            for(; (cjind<cjind1); cjind++)
            {
# 1 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h" 1
# 48 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                {
                    int cj,aj,ajx,ajy,ajz;
# 64 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    __m128 jxSSE,jySSE,jzSSE;
                    __m128 dx_SSE0,dy_SSE0,dz_SSE0;
                    __m128 dx_SSE1,dy_SSE1,dz_SSE1;
                    __m128 dx_SSE2,dy_SSE2,dz_SSE2;
                    __m128 dx_SSE3,dy_SSE3,dz_SSE3;
                    __m128 tx_SSE0,ty_SSE0,tz_SSE0;
                    __m128 tx_SSE1,ty_SSE1,tz_SSE1;
                    __m128 tx_SSE2,ty_SSE2,tz_SSE2;
                    __m128 tx_SSE3,ty_SSE3,tz_SSE3;
                    __m128 rsq_SSE0,rinv_SSE0,rinvsq_SSE0;
                    __m128 rsq_SSE1,rinv_SSE1,rinvsq_SSE1;
                    __m128 rsq_SSE2,rinv_SSE2,rinvsq_SSE2;
                    __m128 rsq_SSE3,rinv_SSE3,rinvsq_SSE3;
                    __m128 wco_SSE0;
                    __m128 wco_SSE1;
                    __m128 wco_SSE2;
                    __m128 wco_SSE3;

                    __m128 jq_SSE;
                    __m128 qq_SSE0;
                    __m128 qq_SSE1;
                    __m128 qq_SSE2;
                    __m128 qq_SSE3;

                    __m128 fexcl_SSE0;
                    __m128 fexcl_SSE1;
                    __m128 fexcl_SSE2;
                    __m128 fexcl_SSE3;

                    __m128 frcoul_SSE0;
                    __m128 frcoul_SSE1;
                    __m128 frcoul_SSE2;
                    __m128 frcoul_SSE3;

                    __m128 r_SSE0,rs_SSE0,rf_SSE0,frac_SSE0;
                    __m128 r_SSE1,rs_SSE1,rf_SSE1,frac_SSE1;
                    __m128 r_SSE2,rs_SSE2,rf_SSE2,frac_SSE2;
                    __m128 r_SSE3,rs_SSE3,rf_SSE3,frac_SSE3;

                    __m128i ti_SSE0,ti_SSE1,ti_SSE2,ti_SSE3;

                    __m128 ctab0_SSE0,ctab1_SSE0;
                    __m128 ctab0_SSE1,ctab1_SSE1;
                    __m128 ctab0_SSE2,ctab1_SSE2;
                    __m128 ctab0_SSE3,ctab1_SSE3;
# 135 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    __m128 fscal_SSE0;
                    __m128 fscal_SSE1;
                    __m128 fscal_SSE2;
                    __m128 fscal_SSE3;
# 160 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    __m128 c6s_j_SSE,c12s_j_SSE;

                    int aj2;

                    __m128 c6_SSE0,c12_SSE0;
                    __m128 c6_SSE1,c12_SSE1;

                    __m128 c6_SSE2,c12_SSE2;
                    __m128 c6_SSE3,c12_SSE3;
# 184 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    __m128 rinvsix_SSE0;
                    __m128 rinvsix_SSE1;

                    __m128 rinvsix_SSE2;
                    __m128 rinvsix_SSE3;



                    __m128 FrLJ6_SSE0,FrLJ12_SSE0;
                    __m128 FrLJ6_SSE1,FrLJ12_SSE1;

                    __m128 FrLJ6_SSE2,FrLJ12_SSE2;
                    __m128 FrLJ6_SSE3,FrLJ12_SSE3;
# 208 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    cj = l_cj[cjind].cj;

                    aj = cj*4;

                    aj2 = aj*2;

                    ajx = aj*3;

                    ajy = ajx + 4;
                    ajz = ajy + 4;
# 258 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    jxSSE = _mm_load_ps(x+ajx);
                    jySSE = _mm_load_ps(x+ajy);
                    jzSSE = _mm_load_ps(x+ajz);


                    dx_SSE0 = _mm_sub_ps(ix_SSE0,jxSSE);
                    dy_SSE0 = _mm_sub_ps(iy_SSE0,jySSE);
                    dz_SSE0 = _mm_sub_ps(iz_SSE0,jzSSE);
                    dx_SSE1 = _mm_sub_ps(ix_SSE1,jxSSE);
                    dy_SSE1 = _mm_sub_ps(iy_SSE1,jySSE);
                    dz_SSE1 = _mm_sub_ps(iz_SSE1,jzSSE);
                    dx_SSE2 = _mm_sub_ps(ix_SSE2,jxSSE);
                    dy_SSE2 = _mm_sub_ps(iy_SSE2,jySSE);
                    dz_SSE2 = _mm_sub_ps(iz_SSE2,jzSSE);
                    dx_SSE3 = _mm_sub_ps(ix_SSE3,jxSSE);
                    dy_SSE3 = _mm_sub_ps(iy_SSE3,jySSE);
                    dz_SSE3 = _mm_sub_ps(iz_SSE3,jzSSE);


                    rsq_SSE0 = gmx_mm_calc_rsq_ps(dx_SSE0,dy_SSE0,dz_SSE0);
                    rsq_SSE1 = gmx_mm_calc_rsq_ps(dx_SSE1,dy_SSE1,dz_SSE1);
                    rsq_SSE2 = gmx_mm_calc_rsq_ps(dx_SSE2,dy_SSE2,dz_SSE2);
                    rsq_SSE3 = gmx_mm_calc_rsq_ps(dx_SSE3,dy_SSE3,dz_SSE3);

                    wco_SSE0 = _mm_cmplt_ps(rsq_SSE0,rc2_SSE);
                    wco_SSE1 = _mm_cmplt_ps(rsq_SSE1,rc2_SSE);
                    wco_SSE2 = _mm_cmplt_ps(rsq_SSE2,rc2_SSE);
                    wco_SSE3 = _mm_cmplt_ps(rsq_SSE3,rc2_SSE);
# 360 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    rinv_SSE0 = gmx_mm_invsqrt_ps(rsq_SSE0);
                    rinv_SSE1 = gmx_mm_invsqrt_ps(rsq_SSE1);
                    rinv_SSE2 = gmx_mm_invsqrt_ps(rsq_SSE2);
                    rinv_SSE3 = gmx_mm_invsqrt_ps(rsq_SSE3);

                    jq_SSE = _mm_load_ps(q+aj);
                    qq_SSE0 = _mm_mul_ps(iq_SSE0,jq_SSE);
                    qq_SSE1 = _mm_mul_ps(iq_SSE1,jq_SSE);
                    qq_SSE2 = _mm_mul_ps(iq_SSE2,jq_SSE);
                    qq_SSE3 = _mm_mul_ps(iq_SSE3,jq_SSE);
# 395 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    c6s_j_SSE = _mm_load_ps(ljc+aj2+0);
                    c12s_j_SSE = _mm_load_ps(ljc+aj2+4);
                    c6_SSE0 = _mm_mul_ps(c6s_SSE0 ,c6s_j_SSE );
                    c6_SSE1 = _mm_mul_ps(c6s_SSE1 ,c6s_j_SSE );

                    c6_SSE2 = _mm_mul_ps(c6s_SSE2 ,c6s_j_SSE );
                    c6_SSE3 = _mm_mul_ps(c6s_SSE3 ,c6s_j_SSE );

                    c12_SSE0 = _mm_mul_ps(c12s_SSE0,c12s_j_SSE);
                    c12_SSE1 = _mm_mul_ps(c12s_SSE1,c12s_j_SSE);

                    c12_SSE2 = _mm_mul_ps(c12s_SSE2,c12s_j_SSE);
                    c12_SSE3 = _mm_mul_ps(c12s_SSE3,c12s_j_SSE);
# 427 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    rinv_SSE0 = _mm_and_ps(rinv_SSE0,wco_SSE0);
                    rinv_SSE1 = _mm_and_ps(rinv_SSE1,wco_SSE1);
                    rinv_SSE2 = _mm_and_ps(rinv_SSE2,wco_SSE2);
                    rinv_SSE3 = _mm_and_ps(rinv_SSE3,wco_SSE3);

                    rinvsq_SSE0 = _mm_mul_ps(rinv_SSE0,rinv_SSE0);
                    rinvsq_SSE1 = _mm_mul_ps(rinv_SSE1,rinv_SSE1);
                    rinvsq_SSE2 = _mm_mul_ps(rinv_SSE2,rinv_SSE2);
                    rinvsq_SSE3 = _mm_mul_ps(rinv_SSE3,rinv_SSE3);
# 471 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    r_SSE0 = _mm_mul_ps(rsq_SSE0,rinv_SSE0);
                    r_SSE1 = _mm_mul_ps(rsq_SSE1,rinv_SSE1);
                    r_SSE2 = _mm_mul_ps(rsq_SSE2,rinv_SSE2);
                    r_SSE3 = _mm_mul_ps(rsq_SSE3,rinv_SSE3);

                    rs_SSE0 = _mm_mul_ps(r_SSE0,invtsp_SSE);
                    rs_SSE1 = _mm_mul_ps(r_SSE1,invtsp_SSE);
                    rs_SSE2 = _mm_mul_ps(r_SSE2,invtsp_SSE);
                    rs_SSE3 = _mm_mul_ps(r_SSE3,invtsp_SSE);

                    ti_SSE0 = _mm_cvttps_epi32(rs_SSE0);
                    ti_SSE1 = _mm_cvttps_epi32(rs_SSE1);
                    ti_SSE2 = _mm_cvttps_epi32(rs_SSE2);
                    ti_SSE3 = _mm_cvttps_epi32(rs_SSE3);

                    rf_SSE0 = _mm_cvtepi32_ps(ti_SSE0);
                    rf_SSE1 = _mm_cvtepi32_ps(ti_SSE1);
                    rf_SSE2 = _mm_cvtepi32_ps(ti_SSE2);
                    rf_SSE3 = _mm_cvtepi32_ps(ti_SSE3);

                    frac_SSE0 = _mm_sub_ps(rs_SSE0,rf_SSE0);
                    frac_SSE1 = _mm_sub_ps(rs_SSE1,rf_SSE1);
                    frac_SSE2 = _mm_sub_ps(rs_SSE2,rf_SSE2);
                    frac_SSE3 = _mm_sub_ps(rs_SSE3,rf_SSE3);


                    { __m128 ctab_SSE[4]; int j; _mm_store_si128((__m128i *)ti0,ti_SSE0); for(j=0; j<4; j++) { ctab_SSE[j] = _mm_load_ps(tab_coul_F+ti0[j]*4); } { __m128 _c01,_c23; _c01 = _mm_shuffle_ps(ctab_SSE[0],ctab_SSE[1],_MM_SHUFFLE(1,0,1,0)); _c23 = _mm_shuffle_ps(ctab_SSE[2],ctab_SSE[3],_MM_SHUFFLE(1,0,1,0)); ctab0_SSE0 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(2,0,2,0)); ctab1_SSE0 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(3,1,3,1)); }; };
                    { __m128 ctab_SSE[4]; int j; _mm_store_si128((__m128i *)ti1,ti_SSE1); for(j=0; j<4; j++) { ctab_SSE[j] = _mm_load_ps(tab_coul_F+ti1[j]*4); } { __m128 _c01,_c23; _c01 = _mm_shuffle_ps(ctab_SSE[0],ctab_SSE[1],_MM_SHUFFLE(1,0,1,0)); _c23 = _mm_shuffle_ps(ctab_SSE[2],ctab_SSE[3],_MM_SHUFFLE(1,0,1,0)); ctab0_SSE1 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(2,0,2,0)); ctab1_SSE1 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(3,1,3,1)); }; };
                    { __m128 ctab_SSE[4]; int j; _mm_store_si128((__m128i *)ti2,ti_SSE2); for(j=0; j<4; j++) { ctab_SSE[j] = _mm_load_ps(tab_coul_F+ti2[j]*4); } { __m128 _c01,_c23; _c01 = _mm_shuffle_ps(ctab_SSE[0],ctab_SSE[1],_MM_SHUFFLE(1,0,1,0)); _c23 = _mm_shuffle_ps(ctab_SSE[2],ctab_SSE[3],_MM_SHUFFLE(1,0,1,0)); ctab0_SSE2 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(2,0,2,0)); ctab1_SSE2 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(3,1,3,1)); }; };
                    { __m128 ctab_SSE[4]; int j; _mm_store_si128((__m128i *)ti3,ti_SSE3); for(j=0; j<4; j++) { ctab_SSE[j] = _mm_load_ps(tab_coul_F+ti3[j]*4); } { __m128 _c01,_c23; _c01 = _mm_shuffle_ps(ctab_SSE[0],ctab_SSE[1],_MM_SHUFFLE(1,0,1,0)); _c23 = _mm_shuffle_ps(ctab_SSE[2],ctab_SSE[3],_MM_SHUFFLE(1,0,1,0)); ctab0_SSE3 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(2,0,2,0)); ctab1_SSE3 = _mm_shuffle_ps(_c01,_c23,_MM_SHUFFLE(3,1,3,1)); }; };
# 525 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    fexcl_SSE0 = _mm_add_ps(ctab0_SSE0,_mm_mul_ps(frac_SSE0,ctab1_SSE0));
                    fexcl_SSE1 = _mm_add_ps(ctab0_SSE1,_mm_mul_ps(frac_SSE1,ctab1_SSE1));
                    fexcl_SSE2 = _mm_add_ps(ctab0_SSE2,_mm_mul_ps(frac_SSE2,ctab1_SSE2));
                    fexcl_SSE3 = _mm_add_ps(ctab0_SSE3,_mm_mul_ps(frac_SSE3,ctab1_SSE3));
                    frcoul_SSE0 = _mm_mul_ps(qq_SSE0,_mm_sub_ps(rinv_SSE0,_mm_mul_ps(fexcl_SSE0,r_SSE0)));
                    frcoul_SSE1 = _mm_mul_ps(qq_SSE1,_mm_sub_ps(rinv_SSE1,_mm_mul_ps(fexcl_SSE1,r_SSE1)));
                    frcoul_SSE2 = _mm_mul_ps(qq_SSE2,_mm_sub_ps(rinv_SSE2,_mm_mul_ps(fexcl_SSE2,r_SSE2)));
                    frcoul_SSE3 = _mm_mul_ps(qq_SSE3,_mm_sub_ps(rinv_SSE3,_mm_mul_ps(fexcl_SSE3,r_SSE3)));
# 626 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    rinvsix_SSE0 = _mm_mul_ps(rinvsq_SSE0,_mm_mul_ps(rinvsq_SSE0,rinvsq_SSE0));
                    rinvsix_SSE1 = _mm_mul_ps(rinvsq_SSE1,_mm_mul_ps(rinvsq_SSE1,rinvsq_SSE1));


                    rinvsix_SSE2 = _mm_mul_ps(rinvsq_SSE2,_mm_mul_ps(rinvsq_SSE2,rinvsq_SSE2));
                    rinvsix_SSE3 = _mm_mul_ps(rinvsq_SSE3,_mm_mul_ps(rinvsq_SSE3,rinvsq_SSE3));


                    FrLJ6_SSE0 = _mm_mul_ps(c6_SSE0,rinvsix_SSE0);
                    FrLJ6_SSE1 = _mm_mul_ps(c6_SSE1,rinvsix_SSE1);

                    FrLJ6_SSE2 = _mm_mul_ps(c6_SSE2,rinvsix_SSE2);
                    FrLJ6_SSE3 = _mm_mul_ps(c6_SSE3,rinvsix_SSE3);

                    FrLJ12_SSE0 = _mm_mul_ps(c12_SSE0,_mm_mul_ps(rinvsix_SSE0,rinvsix_SSE0));
                    FrLJ12_SSE1 = _mm_mul_ps(c12_SSE1,_mm_mul_ps(rinvsix_SSE1,rinvsix_SSE1));

                    FrLJ12_SSE2 = _mm_mul_ps(c12_SSE2,_mm_mul_ps(rinvsix_SSE2,rinvsix_SSE2));
                    FrLJ12_SSE3 = _mm_mul_ps(c12_SSE3,_mm_mul_ps(rinvsix_SSE3,rinvsix_SSE3));
# 746 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_inner.h"
                    fscal_SSE0 = _mm_mul_ps(rinvsq_SSE0,
                                            _mm_add_ps(frcoul_SSE0,
                                                       _mm_sub_ps(FrLJ12_SSE0,FrLJ6_SSE0)));

                    fscal_SSE1 = _mm_mul_ps(rinvsq_SSE1,
                                            _mm_add_ps(frcoul_SSE1,
                                                       _mm_sub_ps(FrLJ12_SSE1,FrLJ6_SSE1)));

                    fscal_SSE2 = _mm_mul_ps(rinvsq_SSE2,
                                            _mm_add_ps(frcoul_SSE2,
                                                       _mm_sub_ps(FrLJ12_SSE2,FrLJ6_SSE2)));

                    fscal_SSE3 = _mm_mul_ps(rinvsq_SSE3,
                                            _mm_add_ps(frcoul_SSE3,
                                                       _mm_sub_ps(FrLJ12_SSE3,FrLJ6_SSE3)));


                    tx_SSE0 = _mm_mul_ps(fscal_SSE0,dx_SSE0);
                    tx_SSE1 = _mm_mul_ps(fscal_SSE1,dx_SSE1);
                    tx_SSE2 = _mm_mul_ps(fscal_SSE2,dx_SSE2);
                    tx_SSE3 = _mm_mul_ps(fscal_SSE3,dx_SSE3);
                    ty_SSE0 = _mm_mul_ps(fscal_SSE0,dy_SSE0);
                    ty_SSE1 = _mm_mul_ps(fscal_SSE1,dy_SSE1);
                    ty_SSE2 = _mm_mul_ps(fscal_SSE2,dy_SSE2);
                    ty_SSE3 = _mm_mul_ps(fscal_SSE3,dy_SSE3);
                    tz_SSE0 = _mm_mul_ps(fscal_SSE0,dz_SSE0);
                    tz_SSE1 = _mm_mul_ps(fscal_SSE1,dz_SSE1);
                    tz_SSE2 = _mm_mul_ps(fscal_SSE2,dz_SSE2);
                    tz_SSE3 = _mm_mul_ps(fscal_SSE3,dz_SSE3);


                    fix_SSE0 = _mm_add_ps(fix_SSE0,tx_SSE0);
                    fix_SSE1 = _mm_add_ps(fix_SSE1,tx_SSE1);
                    fix_SSE2 = _mm_add_ps(fix_SSE2,tx_SSE2);
                    fix_SSE3 = _mm_add_ps(fix_SSE3,tx_SSE3);
                    fiy_SSE0 = _mm_add_ps(fiy_SSE0,ty_SSE0);
                    fiy_SSE1 = _mm_add_ps(fiy_SSE1,ty_SSE1);
                    fiy_SSE2 = _mm_add_ps(fiy_SSE2,ty_SSE2);
                    fiy_SSE3 = _mm_add_ps(fiy_SSE3,ty_SSE3);
                    fiz_SSE0 = _mm_add_ps(fiz_SSE0,tz_SSE0);
                    fiz_SSE1 = _mm_add_ps(fiz_SSE1,tz_SSE1);
                    fiz_SSE2 = _mm_add_ps(fiz_SSE2,tz_SSE2);
                    fiz_SSE3 = _mm_add_ps(fiz_SSE3,tz_SSE3);


                    _mm_store_ps(f+ajx,
                                 _mm_sub_ps( _mm_load_ps(f+ajx), gmx_mm_sum4_ps(tx_SSE0,tx_SSE1,tx_SSE2,tx_SSE3) ));
                    _mm_store_ps(f+ajy,
                                 _mm_sub_ps( _mm_load_ps(f+ajy), gmx_mm_sum4_ps(ty_SSE0,ty_SSE1,ty_SSE2,ty_SSE3) ));
                    _mm_store_ps(f+ajz,
                                 _mm_sub_ps( _mm_load_ps(f+ajz), gmx_mm_sum4_ps(tz_SSE0,tz_SSE1,tz_SSE2,tz_SSE3) ));
                }
# 616 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h" 2
            }

        }
# 653 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
        ninner += cjind1 - cjind0;
# 666 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
        { _MM_TRANSPOSE4_PS(fix_SSE0,fix_SSE1,fix_SSE2,fix_SSE3); fix_SSE0 = _mm_add_ps(fix_SSE0,fix_SSE1); fix_SSE2 = _mm_add_ps(fix_SSE2,fix_SSE3); fix_SSE = _mm_add_ps(fix_SSE0,fix_SSE2); };
        _mm_store_ps(f+scix, _mm_add_ps(fix_SSE, _mm_load_ps(f+scix)));

        { _MM_TRANSPOSE4_PS(fiy_SSE0,fiy_SSE1,fiy_SSE2,fiy_SSE3); fiy_SSE0 = _mm_add_ps(fiy_SSE0,fiy_SSE1); fiy_SSE2 = _mm_add_ps(fiy_SSE2,fiy_SSE3); fiy_SSE = _mm_add_ps(fiy_SSE0,fiy_SSE2); };
        _mm_store_ps(f+sciy, _mm_add_ps(fiy_SSE, _mm_load_ps(f+sciy)));

        { _MM_TRANSPOSE4_PS(fiz_SSE0,fiz_SSE1,fiz_SSE2,fiz_SSE3); fiz_SSE0 = _mm_add_ps(fiz_SSE0,fiz_SSE1); fiz_SSE2 = _mm_add_ps(fiz_SSE2,fiz_SSE3); fiz_SSE = _mm_add_ps(fiz_SSE0,fiz_SSE2); };
        _mm_store_ps(f+sciz, _mm_add_ps(fiz_SSE, _mm_load_ps(f+sciz)));


        _mm_store_ps(shf,fix_SSE);
        fshift[ish3+0] += (shf[0]+shf[1]+shf[2]+shf[3]);
        _mm_store_ps(shf,fiy_SSE);
        fshift[ish3+1] += (shf[0]+shf[1]+shf[2]+shf[3]);
        _mm_store_ps(shf,fiz_SSE);
        fshift[ish3+2] += (shf[0]+shf[1]+shf[2]+shf[3]);
# 721 "/home/pszilard/projects/gmx/gromacs-gpu/src/mdlib/nbnxn_kernels/nbnxn_kernel_sse_outer.h"
    }
}


typedef void (*p_nbk_func_ener)(const nbnxn_pairlist_t     *nbl,
                                const nbnxn_atomdata_t     *nbat,
                                const interaction_const_t  *ic,
                                rvec                       *shift_vec,
                                real                       *f,
                                real                       *fshift,
                                real                       *Vvdw,
                                real                       *Vc);

typedef void (*p_nbk_func_noener)(const nbnxn_pairlist_t     *nbl,
                                  const nbnxn_atomdata_t     *nbat,
                                  const interaction_const_t  *ic,
                                  rvec                       *shift_vec,
                                  real                       *f,
                                  real                       *fshift);

enum { coultRF, coultTAB, coultNR };


p_nbk_func_ener p_nbk_ener[coultNR][ljcrNR] =
{ { nbnxn_kernel_sse_single_rf_comb_geom_ener,
    nbnxn_kernel_sse_single_rf_comb_lb_ener,
    nbnxn_kernel_sse_single_rf_comb_none_ener },
  { nbnxn_kernel_sse_single_tab_comb_geom_ener,
    nbnxn_kernel_sse_single_tab_comb_lb_ener,
    nbnxn_kernel_sse_single_tab_comb_none_ener } };

p_nbk_func_ener p_nbk_energrp[coultNR][ljcrNR] =
{ { nbnxn_kernel_sse_single_rf_comb_geom_energrp,
    nbnxn_kernel_sse_single_rf_comb_lb_energrp,
    nbnxn_kernel_sse_single_rf_comb_none_energrp },
  { nbnxn_kernel_sse_single_tab_comb_geom_energrp,
    nbnxn_kernel_sse_single_tab_comb_lb_energrp,
    nbnxn_kernel_sse_single_tab_comb_none_energrp } };

p_nbk_func_noener p_nbk_noener[coultNR][ljcrNR] =
{ { nbnxn_kernel_sse_single_rf_comb_geom_noener,
    nbnxn_kernel_sse_single_rf_comb_lb_noener,
    nbnxn_kernel_sse_single_rf_comb_none_noener },
  { nbnxn_kernel_sse_single_tab_comb_geom_noener,
    nbnxn_kernel_sse_single_tab_comb_lb_noener,
    nbnxn_kernel_sse_single_tab_comb_none_noener } };


static void clear_f(const nbnxn_atomdata_t *nbat,
                    real *f)
{
    int i;

    for(i=0; i<nbat->natoms*nbat->xstride; i++)
    {
        f[i] = 0;
    }
}

static void clear_fshift(real *fshift)
{
    int i;

    for(i=0; i<SHIFTS*DIM; i++)
    {
        fshift[i] = 0;
    }
}

static void reduce_group_energies(int ng,int ng_2log,
                                  const real *VSvdw,const real *VSc,
                                  real *Vvdw,real *Vc)
{
    int ng_p2,i,j,j0,j1,c,s;

#ifndef GMX_DOUBLE
#define SIMD_WIDTH   4
#define SIMD_WIDTH_2 2
#else
#define SIMD_WIDTH   2
#define SIMD_WIDTH_2 1
#endif

    ng_p2 = (1<<ng_2log);

    /* The size of the SSE energy group buffer array is:
     * stride^3*SIMD_WIDTH_2*SIMD_WIDTH
     */
    for(i=0; i<ng; i++)
    {
        for(j=0; j<ng; j++)
        {
            Vvdw[i*ng+j] = 0;
            Vc[i*ng+j]   = 0;
        }

        for(j1=0; j1<ng; j1++)
        {
            for(j0=0; j0<ng; j0++)
            {
                c = ((i*ng + j1)*ng_p2 + j0)*SIMD_WIDTH_2*SIMD_WIDTH;
                for(s=0; s<SIMD_WIDTH_2; s++)
                {
                    Vvdw[i*ng+j0] += VSvdw[c+0];
                    Vvdw[i*ng+j1] += VSvdw[c+1];
                    Vc  [i*ng+j0] += VSc  [c+0];
                    Vc  [i*ng+j1] += VSc  [c+1];
                    c += SIMD_WIDTH + 2;
                }
            }
        }
    }
}

void
nbnxn_kernel_sse(nbnxn_pairlist_set_t       *nbl_list,
                 const nbnxn_atomdata_t     *nbat,
                 const interaction_const_t  *ic,
                 rvec                       *shift_vec, 
                 int                        force_flags,
                 gmx_bool                   clearF,
                 real                       *fshift,
                 real                       *Vc,
                 real                       *Vvdw)
{
    int              nnbl;
    nbnxn_pairlist_t **nbl;
    int coult;
    int nb;

    nnbl = nbl_list->nnbl;
    nbl  = nbl_list->nbl;

    if (EEL_RF(ic->eeltype) || ic->eeltype == eelCUT)
    {
        coult = coultRF;
    }
    else
    {
        coult = coultTAB;
    }

#pragma omp parallel for schedule(static) num_threads(gmx_omp_nthreads_get(emntNonbonded))
    for(nb=0; nb<nnbl; nb++)
    {
        nbnxn_atomdata_output_t *out;
        real *fshift_p;

        out = &nbat->out[nb];

        if (clearF)
        {
            clear_f(nbat,out->f);
        }

        if ((force_flags & GMX_FORCE_VIRIAL) && nnbl == 1)
        {
            fshift_p = fshift;
        }
        else
        {
            fshift_p = out->fshift;

            if (clearF)
            {
                clear_fshift(fshift_p);
            }
        }

        nbnxn_kernel_sse_single_tab_comb_geom_noener_noLJ_stripped(nbl[nb],nbat,                                                 ic,
                                                                   shift_vec,
                                                                   out->f,
                                                                   fshift_p);

#if 0
        /* With Ewald type electrostatics we the forces for excluded atom pairs
         * should not contribute to the virial sum. The exclusion forces
         * are not calculate in the energy kernels, but are in _noener.
         */
        if (!((force_flags & GMX_FORCE_ENERGY) ||
              (EEL_FULL(ic->eeltype) && (force_flags & GMX_FORCE_VIRIAL))))
        {
            /* Don't calculate energies */
            p_nbk_noener[coult][nbat->comb_rule](nbl[nb],nbat,
                                                 ic,
                                                 shift_vec,
                                                 out->f,
                                                 fshift_p);
        }
        else if (out->nV == 1 || !(force_flags & GMX_FORCE_ENERGY))
        {
            /* No energy groups */
            out->Vvdw[0] = 0;
            out->Vc[0]   = 0;

            p_nbk_ener[coult][nbat->comb_rule](nbl[nb],nbat,
                                               ic,
                                               shift_vec,
                                               out->f,
                                               fshift_p,
                                               out->Vvdw,
                                               out->Vc);
        }
        else
        {
            /* Calculate energy group contributions */
            int i;

            for(i=0; i<out->nVS; i++)
            {
                out->VSvdw[i] = 0;
            }
            for(i=0; i<out->nVS; i++)
            {
                out->VSc[i] = 0;
            }

            p_nbk_energrp[coult][nbat->comb_rule](nbl[nb],nbat,
                                                  ic,
                                                  shift_vec,
                                                  out->f,
                                                  fshift_p,
                                                  out->VSvdw,
                                                  out->VSc);

            reduce_group_energies(nbat->nenergrp,nbat->neg_2log,
                                  out->VSvdw,out->VSc,
                                  out->Vvdw,out->Vc);
        }
#endif /* 0 */
    }

    if (force_flags & GMX_FORCE_ENERGY)
    {
        /* Reduce the energies */
        for(nb=0; nb<nnbl; nb++)
        {
            int i;

            for(i=0; i<nbat->out[nb].nV; i++)
            {
                Vvdw[i] += nbat->out[nb].Vvdw[i];
                Vc[i]   += nbat->out[nb].Vc[i];
            }
        }
        /* printf("vdw %f c %f\n",Vvdw[0],Vc[0]); */
    }
}
